#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>

#include "../include/os.h"

#ifdef __MSW__
# include <windows.h>
#endif
  
#include <GL/gl.h>
#include <GL/glu.h>
         
#include "../include/string.h"  

#include "v3dtex.h"

#include "v3dfio.h"
#include "v3dgl.h"

#include "gw.h"
#include "stategl.h"

#include "image.h"
#include "menu.h"
#include "sar.h"
#include "config.h"


int SARMenuCreateObjView(
        sar_menu_struct *m,
        double x, double y,
        double width, double height,
        GWFont *font,
        sar_menu_color_struct *color,
        const sar_image_struct **bg_image,      /* Pointer to 9 images. */
        const char *title
);
int SARMenuObjViewManagePointer(
        gw_display_struct *display, sar_menu_struct *m,
        sar_menu_objview_struct *ov_ptr, int ov_num,
        int x, int y, int state, int btn_num
);
static int SARMenuObjViewLoadLine(
        gw_display_struct *display, sar_menu_struct *m,
        sar_menu_objview_struct *ov_ptr, int ov_num,
	const char *line
);
int SARMenuObjViewLoad(
        gw_display_struct *display, sar_menu_struct *m,
        sar_menu_objview_struct *ov_ptr, int ov_num,
        const char *filename
);
void SARMenuObjViewRebufferImage(
        gw_display_struct *display, sar_menu_struct *m,
        sar_menu_objview_struct *ov_ptr, int ov_num
);


#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)

#define ISCOMMENT(c)    ((c) == SAR_COMMENT_CHAR)
#define ISCR(c)         (((c) == '\n') || ((c) == '\r'))


/*
 *      Default spacings (in pixels):
 *
 *	These values should match and coenciding values in menu.c.
 */
#define DEF_XMARGIN             5
#define DEF_YMARGIN             5

#define DEF_WIDTH               100
#define DEF_HEIGHT              100


#define DO_REDRAW_MENU		\
{ \
 SARMenuDrawAll(display, m); \
 GWSwapBuffer(display); \
}

#define DO_REDRAW_OBJECT(d,m,n)	\
{ \
 SARMenuDrawObject(display,m,n); \
 GWSwapBuffer(display); \
}



/*
 *	Creates a 3d object viewer object on menu m, returns the index
 *	number of the 3d object viewer object or -1 on failure.
 */
int SARMenuCreateObjView(
        sar_menu_struct *m,
        double x, double y,
        double width, double height,
        GWFont *font,
        sar_menu_color_struct *color,
        const sar_image_struct **bg_image,      /* Pointer to 9 images. */
        const char *title
)
{
	int i, n;
        sar_menu_objview_struct *ov_ptr;


        if(m == NULL)
            return(-1);

        if(m->total_objects < 0)
            m->total_objects = 0;

        for(i = 0; i < m->total_objects; i++)
        {
            if(m->object[i] == NULL)
                break;
        }
        if(i < m->total_objects)
        {
            n = i;
        }
        else
        {
            n = m->total_objects;
            m->total_objects++;

            m->object = (void **)realloc(
                m->object,
                m->total_objects * sizeof(void *)
            );
            if(m->object == NULL)
            {
                m->total_objects = 0;
                return(-1);
            }

            m->object[n] = NULL;
        }   
            
        /* Allocate structure. */
        ov_ptr = (sar_menu_objview_struct *)calloc(
            1, sizeof(sar_menu_objview_struct)
        );
        if(ov_ptr == NULL)
            return(-1);
        else
            m->object[n] = (void *)ov_ptr;


        /* Set 3d object viewer values. */
        ov_ptr->type = SAR_MENU_OBJECT_TYPE_OBJVIEW;
        ov_ptr->x = x;
        ov_ptr->y = y;
        ov_ptr->width = width;
        ov_ptr->height = height;
	ov_ptr->font = font;
	if(color != NULL)
	    memcpy(&ov_ptr->color, color, sizeof(sar_menu_color_struct));
	ov_ptr->bg_image = bg_image;
	ov_ptr->title = (title != NULL) ? strdup(title) : NULL;
	ov_ptr->objlist = NULL;
	ov_ptr->objheading = 0.0 * PI;
        ov_ptr->objpitch = 0.0 * PI;
        ov_ptr->objbank = 0.0 * PI;
	ov_ptr->camera_distance = 20.0;
	ov_ptr->objimg = NULL;
	ov_ptr->objimg_width = 0;
	ov_ptr->objimg_height = 0;

	return(n);
}


/*
 *	Manages the pointer event with respect to the given menu 3d
 *	object viewer object.
 */
int SARMenuObjViewManagePointer(
        gw_display_struct *display, sar_menu_struct *m,
	sar_menu_objview_struct *ov_ptr, int ov_num,
        int x, int y, int state, int btn_num
)
{
        int events_handled = 0;
        int x_min, x_max, y_min, y_max, width, height, w, h;

        static Boolean rotate_drag_state = False;
	static Boolean bank_drag_state = False;
        static int prev_motion_x, prev_motion_y;
        static time_t last_motion_time = 0;


        if((display == NULL) || (m == NULL) || (ov_ptr == NULL))
	    return(events_handled);

        GWContextGet(
            display, 0,
            NULL, NULL,
            NULL, NULL,
            &width, &height
        );

        /* If button states down and event is a release, reset all
         * states.
         */
        if((rotate_drag_state || bank_drag_state) &&
           (state == GWEventTypeButtonRelease)
        )
        {
            rotate_drag_state = False;
	    bank_drag_state = False;
        }

	w = (int)(ov_ptr->width * width);
        if(w <= 0)
            w = (int)(width - (2 * DEF_XMARGIN));
        if(w <= 0)
            w = DEF_XMARGIN;

        if(ov_ptr->height > 0)
            h = (int)(ov_ptr->height * height);
        else
            h = DEF_HEIGHT;
        if(h <= 0)
            h = DEF_YMARGIN;

        /* Get coordinate bounds. */
        x_min = (int)((ov_ptr->x * width) - (w / 2));
        x_max = (int)(x_min + w);
        y_min = (int)((ov_ptr->y * height) - (h / 2));
        y_max = (int)(y_min + h);

        /* Pointer event in bounds? */
        if((x >= x_min) && (x < x_max) &&
           (y >= y_min) && (y < y_max)
        )
        {
	    /* Handle by event type. */
	    switch(state)
	    {
	      case GWEventTypeButtonPress:
                events_handled++;
                m->selected_object = ov_num;	/* Select this object. */

                switch(btn_num)
		{
                  case 1:
                    rotate_drag_state = True;
                    last_motion_time = cur_millitime;
                    prev_motion_x = x;
                    prev_motion_y = y;

                    GWSetPointerCursor(
                        display, GWPointerCursorTranslate
                    );
                    DO_REDRAW_OBJECT(display, m, ov_num);
                    break;

		  case 3:
		    bank_drag_state = True;
                    last_motion_time = cur_millitime;
                    prev_motion_x = x;
                    prev_motion_y = y;

                    GWSetPointerCursor(
                        display, GWPointerCursorTranslate
                    );
                    DO_REDRAW_OBJECT(display, m, ov_num);
                    break;
                }
                break;

	      case GWEventTypeButtonRelease:
                GWSetPointerCursor(
                    display, GWPointerCursorStandard
                );

                /* Manipulate last motion time so that the following
	         * case checks it properly.
                 */
                last_motion_time = cur_millitime - 1;

		/* Fall through. */
	      case GWEventTypePointerMotion:
                /* Too soon to handle a motion event? */
                if(last_motion_time == cur_millitime)
                    break;
                else
                    last_motion_time = cur_millitime;
                events_handled++;

                /* Rotating? */
                if(rotate_drag_state)
                {
		    int dx = x - prev_motion_x,
			dy = y - prev_motion_y;
		    double movement_coeff;


		    if(width > 0)
			movement_coeff = (double)dx / (double)width;
		    else
			movement_coeff = 0.0;

		    ov_ptr->objheading += movement_coeff * 2.0 * PI;
		    while(ov_ptr->objheading >= (2.0 * PI))
			ov_ptr->objheading -= 2.0 * PI;
                    while(ov_ptr->objheading < (0.0 * PI))
                        ov_ptr->objheading += 2.0 * PI;


                    if(height > 0)
                        movement_coeff = (double)dy / (double)height;
                    else
                        movement_coeff = 0.0;

                    ov_ptr->objpitch += movement_coeff * 2.0 * PI;
                    while(ov_ptr->objpitch >= (2.0 * PI))
                        ov_ptr->objpitch -= 2.0 * PI;
                    while(ov_ptr->objpitch < (0.0 * PI))
                        ov_ptr->objpitch += 2.0 * PI;

		    /* Redraw 3d object and regenerate buffered object
		     * image.
		     */
		    SARMenuObjViewRebufferImage(
			display, m, ov_ptr, ov_num
		    );

		    /* Need to redraw everything, since the regeneration
		     * of the buffered object would have messed with
		     * the back buffer and thus the surrounding graphics.
		     *
		     * So simply redrawing the menu object is not
		     * enough to cover all the modified fragments.
		     */
		    GWPostRedraw(display);
                }
		/* Bank? */
		else if(bank_drag_state)
		{
                    int dx = x - prev_motion_x;
                    double movement_coeff;


                    if(width > 0)
                        movement_coeff = (double)dx / (double)width;
                    else
                        movement_coeff = 0.0;

                    ov_ptr->objbank += movement_coeff * 2.0 * PI;
                    while(ov_ptr->objbank >= (2.0 * PI))
                        ov_ptr->objbank -= 2.0 * PI;
                    while(ov_ptr->objbank < (0.0 * PI))
                        ov_ptr->objbank += 2.0 * PI;


                    /* Redraw 3d object and regenerate buffered object
                     * image.
                     */
                    SARMenuObjViewRebufferImage(
                        display, m, ov_ptr, ov_num
                    );

                    /* Need to redraw everything, since the regeneration
                     * of the buffered object would have messed with
                     * the back buffer and thus the surrounding graphics.
                     *
                     * So simply redrawing the menu object is not
                     * enough to cover all the modified fragments.
                     */
                    GWPostRedraw(display);
                }

                /* Record previous pointer coordinates. */
                prev_motion_x = x;
                prev_motion_y = y;
                break;

            }	/* Handle by event type. */

	}	/* Pointer event in bounds? */


	return(events_handled);
}

/*
 *	Called by SARMenuObjViewLoad() to process each line found in
 *	non v3d models in the loaded model file.
 *
 *	This is to get the size of the object (the contact bounds)
 *	the name and description.
 *
 *	Inputs assumed valid.
 *
 *	Returns non-zero on error.
 */
static int SARMenuObjViewLoadLine(
        gw_display_struct *display, sar_menu_struct *m,
        sar_menu_objview_struct *ov_ptr, int ov_num,
        const char *line
)
{
	const char *arg;
	char *strptr;
	char parm[256];


	if(line == NULL)
	    return(-1);

	/* Seek past spaces. */
	while(ISBLANK(*line))
	    line++;

	/* Skip comments. */
	if(ISCOMMENT(*line))
	    return(0);

	/* Get parameter. */
	strncpy(parm, line, 256);
	parm[256 - 1] = '\0';
	strptr = strchr(parm, ' ');
	if(strptr != NULL)
	    *strptr = '\0';
        strptr = strchr(parm, '\t');
        if(strptr != NULL)
            *strptr = '\0';

	/* Get pointer to argument. */
	arg = line;
	while(!ISBLANK(*arg) && (*arg != '\0'))
            arg++;
	while(ISBLANK(*arg))
            arg++;


	/* Name. */
	if(!strcasecmp(parm, "name"))
        {
	    free(ov_ptr->title);
	    ov_ptr->title = strdup(arg);
	}
        /* Description. */
        else if(!strcasecmp(parm, "desc") ||
                !strcasecmp(parm, "description")
	)
        {
            free(ov_ptr->message);
            ov_ptr->message = strdup(arg);
        }
	/* Contact bounds, spherical shape. */
	else if(!strcasecmp(parm, "contact_spherical"))
	{
	    /* Arguments are: <radius>
	     */
	    int sr, sc, st = 1;
	    double tan_value, value[1];
	    sr = sscanf(arg, "%lf", &value[0]);
	    for(sc = sr; sc < st; sc++)
		value[sc] = 0.0;

	    /* Calculate the distance of the camera to be able to
	     * capture the entire size of the object as indicated by
	     * the radius of the contact bounds plus 10% more.
	     *
	     * The field of view will be assumed to be 40 degrees.
	     */
	    tan_value = tan(DEGTORAD(20.0));
	    ov_ptr->camera_distance = (value[0] * 1.1) /
		((tan_value > 0.0) ? tan_value : 1.0);
	}
	/* Contact bounds, cylendrical shape. */
	else if(!strcasecmp(parm, "contact_cylendrical"))
	{
	    /* Arguments are: <radius> <height_min> <height_max>
	     */
	    int sr, sc, st = 3;
	    double tan_value, value[3];
	    sr = sscanf(arg, "%lf %lf %lf",
		&value[0], &value[1], &value[2]
	    );
	    for(sc = sr; sc < st; sc++)
		value[sc] = 0.0;

            /* Calculate the distance of the camera to be able to
             * capture the entire size of the object as indicated by
             * the radius of the contact bounds plus 10% more.
             *
             * The field of view will be assumed to be 40 degrees.
             */
            tan_value = tan(DEGTORAD(20.0));
            ov_ptr->camera_distance = (value[0] * 1.1) /
                ((tan_value > 0.0) ? tan_value : 1.0);
	}

	return(0);
}

/*
 *	Loads a v3d model specified by filename to the 3d object
 *	viewer object's data (deleting the previous model stored on
 *	the object's gl lists as needed).
 *
 *	Returns non-zero on error.
 */
int SARMenuObjViewLoad(
        gw_display_struct *display, sar_menu_struct *m,
        sar_menu_objview_struct *ov_ptr, int ov_num,
        const char *filename
)
{
	FILE *fp;
	GLuint list;
	int i, n, status;
	const char *name;
	void **mh_item = NULL;
	int total_mh_items = 0;
	v3d_model_struct **model = NULL, *model_ptr;
	int total_models = 0;
	v3d_glinterprite_struct *gli;
	v3d_glresource_struct *glres;


	if((display == NULL) || (m == NULL) || (ov_ptr == NULL))
	    return(-1);

	if(filename == NULL)
	    return(-1);


	/* Delete previous gl list of 3d object as needed. */
	list = (GLuint)ov_ptr->objlist;
	if(list != 0)
	{
	    glDeleteLists(list, 1);
	    ov_ptr->objlist = NULL;
	}
	/* Delete previous v3d gl resources for the gl list of
	 * the 3d object as needed.
	 */
	V3DGLResourceDelete(
	    (v3d_glresource_struct *)ov_ptr->objglres
	);
	ov_ptr->objglres = NULL;

	/* Reset object orientation and camera distance. */
	ov_ptr->objheading = 0.0 * PI;
        ov_ptr->objpitch = 0.0 * PI;
        ov_ptr->objbank = 0.0 * PI;
        ov_ptr->camera_distance = 20.0;


	/* Open v3d model file for reading. */
	fp = fopen(filename, "rb");
	if(fp == NULL)
	    return(-1);

	/* Load data from v3d model file. */
	status = V3DLoadModel(
	    NULL, fp,
	    &mh_item, &total_mh_items,
	    &model, &total_models,
	    NULL, NULL
	);

	/* Close v3d model file, data already loaded. */
	fclose(fp);
	fp = NULL;

	/* Error loading v3d model file? */
	if(status)
	    return(-1);

	/* Set up v3d to gl interpritation parameters. */
	gli = V3DGLInterpriteNew();
	if(gli != NULL)
	{
	    gli->flags = V3D_GLFLAG_COORDINATE_AXIS |
		V3D_GLFLAG_TEXTURE_KEEP | V3D_GLFLAG_ALLOW_TRANSLATIONS |
		V3D_GLFLAG_ALLOW_ROTATIONS | V3D_GLFLAG_FLIP_WINDING |
		V3D_GLFLAG_PASS_NORMALS | V3D_GLFLAG_UNITLIZE_NORMALS |
		V3D_GLFLAG_PASS_TEXCOORDS |
		V3D_GLFLAG_TEXTURE_NAME_CASE_SENSITIVE |
		V3D_GLFLAG_MATERIAL_PROPERTIES | V3D_GLFLAG_FACES |
		V3D_GLFLAG_ENABLE_BLENDING | V3D_GLFLAG_SET_BLEND_FUNC |
		V3D_GLFLAG_HEIGHTFIELD_BASE_DIR |
#ifdef __MSW__
		V3D_GLFLAG_TEXTURE_BASE_DIR
#else
		0
#endif
		;

	    gli->coordinate_axis = V3D_GLCOORDINATE_AXIS_SCIENTIFIC;
	    gli->texture_keep = V3D_GLTEXTURE_KEEP_ALWAYS;
	    gli->allow_translations = 1;
	    gli->allow_rotations = 1;
	    gli->flip_winding = 1;
	    gli->pass_normals = V3D_GLPASS_NORMALS_AS_NEEDED;
	    gli->unitlize_normals = 0;
	    gli->pass_texcoords = V3D_GLPASS_TEXCOORDS_AS_NEEDED;
	    gli->texture_name_case_sensitive = 0;
	    gli->material_properties = V3D_GLPASS_MATERIAL_PROPERTIES_WITH_COLOR;
	    gli->faces = V3D_GLFACES_FRONT;
	    gli->enable_blending = V3D_GLENABLE_BLENDING_AS_NEEDED;
	    gli->set_blend_func = 1;
	    gli->heightfield_base_dir = NULL;
#ifdef __MSW__
	    gli->texture_base_dir = strdup(dname.local_data);
#else
	    gli->texture_base_dir = NULL;
#endif

	}

	/* Create a new gl resources from the loaded model data. */
	glres = V3DGLResourceNewFromModelData(
	    gli, mh_item, total_mh_items, model, total_models
	);
	ov_ptr->objglres = (void *)glres;



        /* Create a new gl list. */
        list = glGenLists(1);
	ov_ptr->objlist = (void *)list;

	/* Begin recording gl list. */
	glNewList(list, GL_COMPILE);

	/* Generate gl commands from the loaded v3d models, these gl
	 * commands will be recorded on the gl list.
	 */
	for(i = 0; i < total_models; i++)
	{
	    model_ptr = model[i];
	    if(model_ptr == NULL)
		continue;

	    /* Forward any other data lines found on this model to
	     * our handler.
	     */
	    for(n = 0; n < model_ptr->total_other_data_lines; n++)
		SARMenuObjViewLoadLine(
		    display, m, ov_ptr, ov_num,
		    model_ptr->other_data_line[n]
		);

	    /* Skip all models that are not of type
	     * V3D_MODEL_TYPE_STANDARD since they would not contain any
	     * model data.
	     */
	    if(model_ptr->type != V3D_MODEL_TYPE_STANDARD)
		continue;

	    name = model_ptr->name;
	    if(name == NULL)
		continue;

	    /* Accept only models with the appropriate names. */
	    if(strcmp(name, "standard") &&
               strcmp(name, "landing_gear") &&
               strcmp(name, "rotor") &&
               strcmp(name, "door") &&
               strcmp(name, "flap") &&
               strcmp(name, "air_brake") &&
               strcmp(name, "aileron_left") &&
               strcmp(name, "aileron_right") &&
               strcmp(name, "rudder_top") &&
               strcmp(name, "rudder_bottom") &&
               strcmp(name, "elevator") &&
               strcmp(name, "cannard") &&
               strcmp(name, "aileron_elevator_left") &&
               strcmp(name, "aileron_elevator_right")
	    )
		continue;

	    V3DGLProcessModel(glres, model_ptr);
	}

	/* End gl list recording. */
	glEndList();


	/* Deallocate loaded v3d model data. */
	V3DMHListDeleteAll(&mh_item, &total_mh_items);
	V3DModelListDeleteAll(&model, &total_models);

	/* Deallocate the gl interpritation legend structure, it is no
	 * longer needed.
	 */
	V3DGLInterpriteDelete(gli);

	return(0);
}

/*
 *	Redraws the object in the gl lists of the 3d object viewer
 *	object to its own obj_img buffer, deleting the previous
 *	obj_img buffer as needed.
 */
void SARMenuObjViewRebufferImage(
        gw_display_struct *display, sar_menu_struct *m,
        sar_menu_objview_struct *ov_ptr, int ov_num
)
{
	int width, height;
	double view_aspect;
	void *data;
	GLuint list;
	v3d_glresource_struct *glres;
	GLfloat vf[4];


        if((display == NULL) || (m == NULL) || (ov_ptr == NULL))
            return;

	/* Deallocate previous 3d object buffered image as needed. */
	data = ov_ptr->objimg;
	if(data != NULL)
	{
	    free(data);
	    ov_ptr->objimg = data = NULL;
	    ov_ptr->objimg_width = 0;
	    ov_ptr->objimg_height = 0;
	}

	/* Get the gl display list and the v3d gl resources of the 3d
	 * object.
	 */
	list = (GLuint)ov_ptr->objlist;
	glres = (v3d_glresource_struct *)ov_ptr->objglres;
	if((list == 0) || (glres == NULL))
	    return;

	/* Get geometry of rendering context. */
	GWContextGet(
	    display, 0,
	    NULL, NULL,
	    NULL, NULL,
	    &width, &height
	);

	/* Calculate view aspect ratio. */
        if(height > 0)
            view_aspect = (double)width / (double)height;
        else
            view_aspect = 800.0 / 600.0;


	/* Switch to 3d drawing and set up gl states to begin drawing
	 * to the back buffer. Then when the drawing is done, read back
	 * the back buffer to get the new buffered 3d object image.
	 */

	glViewport(0, 0, width, height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(
	    40.0,		/* Field of view z in degrees. */
	    view_aspect,
	    0.1,
	    10000.0
	);
	glMatrixMode(GL_MODELVIEW);

        glLoadIdentity();

	/* Set up camera. */
	gluLookAt(
	    -ov_ptr->camera_distance, 0.0, -0.0,	/* Eye x, z, -y */
	    0.0, 0.0, -0.0,		/* Center x, z, -y */
	    0.0, 1.0, -0.0		/* Up x, z, -y */
	);

	/* Clear frame buffer. */
	glClearColor(
	    1.0, 1.0, 1.0, 0.0
	);
	glClearDepth((GLclampd)1.0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* Reset depth states. */
	StateGLEnable(&display->state_gl, GL_DEPTH_TEST);
	StateGLDepthFunc(&display->state_gl, GL_LEQUAL);

	/* Reset texture states. */
	StateGLEnable(&display->state_gl, GL_TEXTURE_2D);
	StateGLTexEnvI(
	    &display->state_gl,
	    GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE
	);

	/* Reset surface states. */
	StateGLEnable(&display->state_gl, GL_CULL_FACE);
	StateGLFrontFace(&display->state_gl, GL_CCW);
	StateGLShadeModel(&display->state_gl, GL_SMOOTH);

	/* Reset material values (and no color material). */
        StateGLDisable(&display->state_gl, GL_COLOR_MATERIAL);
        vf[0] = 0.2;
        vf[1] = 0.2;
        vf[2] = 0.2;
        vf[3] = 1.0;
        glMaterialfv(GL_FRONT, GL_AMBIENT, vf);
        vf[0] = 0.8;
        vf[1] = 0.8;
        vf[2] = 0.8;
        vf[3] = 1.0;
        glMaterialfv(GL_FRONT, GL_DIFFUSE, vf);
        vf[0] = 0.2;
        vf[1] = 0.2;
        vf[2] = 0.2;
        vf[3] = 1.0;
        glMaterialfv(GL_FRONT, GL_SPECULAR, vf);
        vf[0] = 0.9 * 128.0;
        glMaterialfv(GL_FRONT, GL_SHININESS, vf);
        vf[0] = 0.0;
        vf[1] = 0.0;
        vf[2] = 0.0;
        vf[3] = 1.0;
        glMaterialfv(GL_FRONT, GL_EMISSION, vf);

	/* Set up lighting. */
        vf[0] = 0.1;
        vf[1] = 0.1;
        vf[2] = 0.1;
        vf[3] = 1.0;
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, vf);

        vf[0] = (GLfloat)-1000.0;
        vf[1] = (GLfloat)1000.0;
        vf[2] = (GLfloat)-1000.0;
	vf[3] = 0.0;			/* Directional light. */
        glLightfv(GL_LIGHT0, GL_POSITION, vf);
        vf[0] = 0.4;
        vf[1] = 0.4;
        vf[2] = 0.4;
        vf[3] = 1.0;
        glLightfv(GL_LIGHT0, GL_AMBIENT, vf);
        vf[0] = 1.0;
        vf[1] = 1.0;
        vf[2] = 1.0;
        vf[3] = 1.0;
        glLightfv(GL_LIGHT0, GL_DIFFUSE, vf);
        vf[0] = 1.0;
        vf[1] = 1.0;
        vf[2] = 1.0;
        vf[3] = 1.0;
        glLightfv(GL_LIGHT0, GL_SPECULAR, vf);
	vf[0] = 1.0;
        glLightfv(GL_LIGHT0, GL_CONSTANT_ATTENUATION, vf);
        vf[0] = 0.0;
        glLightfv(GL_LIGHT0, GL_LINEAR_ATTENUATION, vf);
        vf[0] = 0.0;
        glLightfv(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, vf);

	StateGLEnable(&display->state_gl, GL_LIGHTING);
	StateGLEnable(&display->state_gl, GL_LIGHT0);


	/* Enable alpha testing. */
	StateGLEnable(&display->state_gl, GL_ALPHA_TEST);
	StateGLAlphaFunc(&display->state_gl, GL_GREATER, 0.5);


	glPushMatrix();
	{
	    glRotated(
		-RADTODEG(ov_ptr->objheading),
		0.0, 1.0, 0.0
	    );
	    glRotated(
		-RADTODEG(ov_ptr->objpitch),
		1.0, 0.0, 0.0
	    );
	    glRotated(
		-RADTODEG(ov_ptr->objbank),
		0.0, 0.0, 1.0
	    );

	    /* Call object's gl list to issue prerecorded gl drawing
	     * commands.
	     */
	    glCallList(list);
	}
	glPopMatrix();


	if((width > 0) && (height > 0))
	{
            /* Read the frame buffer into a tempory allocated buffer and
	     * then copy it (because its flipped) to our final buffer.
	     */
	    void *tmp_data = (void *)malloc(
		width * height * 4 * sizeof(GLubyte)
	    );
	    ov_ptr->objimg = data = (void *)malloc(
                width * height * 4 * sizeof(GLubyte)
            );
            ov_ptr->objimg_width = width;
            ov_ptr->objimg_height = height;

	    /* Read pixels from frame buffer to our tempory buffer. */
	    if(tmp_data != NULL)
	    {
		/* Select frame buffer. */
		if(display->has_double_buffer)
		    glReadBuffer(GL_BACK);
		else
		    glReadBuffer(GL_FRONT);

		/* Read pixels from frame buffer. */
		glFlush();
		glReadPixels(
		    0, 0,
		    width, height,	/* Read same size as frame buffer. */
		    GL_RGBA, GL_UNSIGNED_BYTE,
		    (GLvoid *)tmp_data
		);

		/* The read pixels stored in tmp_data are flipped,
		 * so we need to flip them back.
		 */
		if(data != NULL)
		{
		    int y;
		    GLubyte *src = (GLubyte *)tmp_data, *src_ptr;
		    GLubyte *tar = (GLubyte *)data, *tar_ptr;

		    for(y = 0; y < height; y++)
		    {
			src_ptr = &src[
			    y * width * 4 * sizeof(GLubyte)
			];
			tar_ptr = &tar[
			    (height - 1 - y) * width * 4 * sizeof(GLubyte)
			];
			memcpy(
			    tar_ptr, src_ptr,
			    width * 4 * sizeof(GLubyte)
			);
		    }
		}

		/* No longer need tmp_data. */
		free(tmp_data);
	    }
	}


	/* Reset gl states and transform model view matrices for
	 * 2d menu drawing due to the modifications of the gl states
	 * and matrices above.
	 */
	SARMenuGLStateReset(display);
}
