#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <math.h>

#include "../include/os.h"

#include "sound.h"
#include "obj.h"
#include "sarsound.h"
#include "sar.h"
#include "config.h"


#ifdef __MSW__
static double rint(double x);
#endif	/* __MSW__ */


void SARSoundEngineVolumeMute(
        snd_recorder_struct *recorder,
        snd_play_struct *inside_snd_play,
        snd_play_struct *outside_snd_play 
);
void SARSoundAdjustEngineVolume(
        snd_recorder_struct *recorder,
	sar_sound_source_struct **list, int total,
        snd_play_struct *inside_snd_play,
	int inside_sndsrc_num,
        snd_play_struct *outside_snd_play,
        int outside_sndsrc_num,
        int ear_in_cockpit,
        int engine_state,
        double throttle,
        double distance_to_camera
);
void SARSoundStallUpdate(
        snd_recorder_struct *recorder,
        sar_sound_source_struct **list, int total,
	void **stall_snd_play,
        int stall_sndsrc_num,
        int ear_in_cockpit,
        int flight_model, int is_landed,
        double speed, double speed_stall
);
void SARSoundOverSpeedUpdate(
	snd_recorder_struct *recorder,
        sar_sound_source_struct **list, int total,
        void **overspeed_snd_play,
        int overspeed_sndsrc_num,
	int ear_in_cockpit, int is_overspeed
);


#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))


#ifdef __MSW__
static double rint(double x)
{
	if((double)((double)x - (int)x) > (double)0.5)
	    return((double)((int)x + (int)1));
	else
	    return((double)((int)x));
}
#endif	/* __MSW__ */



/*
 *	Basically calls SARSoundAdjustEngineVolume() with no throttle
 *	and engine state as SAR_ENGINE_STATE_OFF.
 */
void SARSoundEngineVolumeMute(
        snd_recorder_struct *recorder,
        snd_play_struct *inside_snd_play,
        snd_play_struct *outside_snd_play 
)
{
	SARSoundAdjustEngineVolume(
	    recorder,
	    NULL, 0,			/* No sound sources needed. */
	    inside_snd_play,
	    -1,
	    outside_snd_play,
	    -1,
	    0,				/* In cockpit? */
	    SAR_ENGINE_STATE_OFF,	/* Engine state? */
	    0.0,			/* Throttle? */
	    0.0
	);
}

/* 
 *      Adjusts the engine volume and sample rate.
 */
void SARSoundAdjustEngineVolume(
        snd_recorder_struct *recorder,
        sar_sound_source_struct **list, int total,
        snd_play_struct *inside_snd_play,
        int inside_sndsrc_num,
        snd_play_struct *outside_snd_play,
        int outside_sndsrc_num,
	int ear_in_cockpit,
        int engine_state,			/* One of SAR_ENGINE_STATE_*. */
        double throttle,			/* Throttle coeff 0.0 to 1.0. */
        double distance_to_camera		/* In meters. */
)
{
        int sample_rate, new_sample_rate, sample_rate_limit;	/* In Hz. */
        double volume;				/* 0.0 to 1.0. */
        double audiable_radius = 2000.0;
	sar_sound_source_struct *inside_sndsrc = NULL,
				*outside_sndsrc = NULL;

        /* Not connected to sound server? */
        if(recorder == NULL)
            return;

        /* Turn all sounds to 0 volume if engine and throttle are off,
	 * this improves efficiency since no sound would be mixed by
	 * the sound server if volume is 0 but the sound object will
	 * still remain valid.
	 */
/*
        if((engine_state != SAR_ENGINE_STATE_ON) &&
           (throttle <= 0.0)
	)
 */
	if(engine_state != SAR_ENGINE_STATE_ON)
        {
            SoundChangePlayVolume(
                recorder, inside_snd_play,
                0.0, 0.0
            );
            SoundChangePlayVolume(
                recorder, outside_snd_play,
                0.0, 0.0
            );
            return;
        }


	/* Get pointers to sound source structures (if possible). */
	if(list != NULL)
	{
	    if((inside_sndsrc_num >= 0) &&
	       (inside_sndsrc_num < total)
	    )
		inside_sndsrc = list[inside_sndsrc_num];

            if((outside_sndsrc_num >= 0) &&
               (outside_sndsrc_num < total)
            )
                outside_sndsrc = list[outside_sndsrc_num];
	}


        /* Begin calculating sample rate based on the throttle position,
	 * increase to a faster sample rate as throttle increases.
         */

	/* Get lower and upper sample rate bounds and ranges. */
	sample_rate_limit = sample_rate = recorder->sample_rate;
	if(ear_in_cockpit && inside_sndsrc != NULL)
	{
	    sample_rate_limit = inside_sndsrc->sample_rate_limit;
	    audiable_radius = inside_sndsrc->range;
	}
	else if(outside_sndsrc != NULL)
	{
	    sample_rate_limit = outside_sndsrc->sample_rate_limit;
	    audiable_radius = outside_sndsrc->range;
	}

	/* Calculate new sample rate based on throttle position. */
	new_sample_rate = (int)rint(
	    ((MAX(sample_rate_limit - sample_rate, 0) * throttle) +
	    sample_rate) / 1000
	);
	new_sample_rate = (int)(new_sample_rate * 1000);


        /* Calculate volume based on (1 - (x^0.5)) curve. */
        volume = CLIP(1.0 -
            pow(distance_to_camera / audiable_radius, 0.5),
            0.0, 1.0
        );

        /* Is ear inside cockpit? */
        if(ear_in_cockpit)
        {
            /* Inside cockpit. */
            SoundChangePlayVolume(
                recorder, inside_snd_play,
                volume, volume
            );
            SoundChangePlaySampleRate(
                recorder, inside_snd_play,
                new_sample_rate
            );
            SoundChangePlayVolume(
                recorder, outside_snd_play,
                0.0, 0.0
            );
            SoundChangePlaySampleRate(
                recorder, outside_snd_play,
                new_sample_rate
            );
        }
        else
        {
            /* Outside. */
            SoundChangePlayVolume(
                recorder, inside_snd_play,
                0.0, 0.0
            );
            SoundChangePlaySampleRate(
                recorder, inside_snd_play,
                new_sample_rate
            );
            SoundChangePlayVolume(
                recorder, outside_snd_play,
                volume, volume
            );
            SoundChangePlaySampleRate(
                recorder, outside_snd_play,
                new_sample_rate
            );
        }
}

/*
 *	Plays or stops the repeating stall warning sound.
 *
 *	The speed_stall should already have any modifications (ie flaps)
 *	applied to it.
 */
void SARSoundStallUpdate(
        snd_recorder_struct *recorder,
        sar_sound_source_struct **list, int total,
	void **stall_snd_play,
        int stall_sndsrc_num,
        int ear_in_cockpit,
        int flight_model, int is_landed,
        double speed, double speed_stall
)
{
	int stall_on = 0;
	double speed_threshold;


        if((recorder == NULL) || (stall_snd_play == NULL))
            return;

	/* Calculate the speed at which the sound should be turned on
	 * if the speed is lower than this value.
	 */
	speed_threshold = speed_stall * 0.8;

	/* Begin checking if stall should be turned on. */
	if(!is_landed && (flight_model == SAR_FLIGHT_MODEL_AIRPLANE))
	{
	    if((speed < speed_threshold) && (speed_threshold > 0.0))
		stall_on = 1;
	}

	/* Check if the sound needs to be played or stopped? */
	if(stall_on && (*stall_snd_play == NULL))
	{
	    /* Need to start playing stall sound. */
	    int i = stall_sndsrc_num;
	    sar_sound_source_struct *sndsrc;


	    if((i >= 0) && (i < total))
		sndsrc = list[i];
	    else
		sndsrc = NULL;

	    if(sndsrc != NULL)
		*stall_snd_play = (void *)SoundStartPlay(
		    recorder,
		    sndsrc->filename,
		    1.0, 1.0,
		    0,
		    SND_PLAY_OPTION_REPEATING
		);
	}
	else if(!stall_on && (*stall_snd_play != NULL))
	{
	    /* Need to stop playing stall sound. */
	    SoundStopPlay(
		recorder,
		(snd_play_struct *)(*stall_snd_play)
	    );
	    *stall_snd_play = NULL;
	}
}

/*
 *      Plays or stops the repeating overspeed warning sound.
 */
void SARSoundOverSpeedUpdate(
        snd_recorder_struct *recorder,
        sar_sound_source_struct **list, int total,
        void **overspeed_snd_play,
        int overspeed_sndsrc_num,
        int ear_in_cockpit, int is_overspeed
)
{
        if((recorder == NULL) || (overspeed_snd_play == NULL))
            return;

        /* Check if the sound needs to be played or stopped? */
        if(is_overspeed && (*overspeed_snd_play == NULL))
        {
            /* Need to start playing overspeed sound. */
            int i = overspeed_sndsrc_num;
            sar_sound_source_struct *sndsrc;


            if((i >= 0) && (i < total))
                sndsrc = list[i];
            else
                sndsrc = NULL;

            if(sndsrc != NULL)
                *overspeed_snd_play = (void *)SoundStartPlay(
                    recorder,
                    sndsrc->filename,
                    1.0, 1.0,
                    0,
                    SND_PLAY_OPTION_REPEATING
                );
        }
        else if(!is_overspeed && (*overspeed_snd_play != NULL))
        {
            /* Need to stop playing overspeed sound. */
            SoundStopPlay(
                recorder,
                (snd_play_struct *)(*overspeed_snd_play)
            );
            *overspeed_snd_play = NULL;
        }
}
